'''多尺度特征+三维直接相减后拉直'''

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Variable
from torch.optim.lr_scheduler import StepLR
import numpy as np
import task_generator as tg
import os
import math
import argparse
import scipy as sp
import scipy.stats
from storage import save_statistics, build_experiment_folder

parser = argparse.ArgumentParser(description="One Shot Visual Recognition")
parser.add_argument("-r1", "--relation1_dim", type=int, default=10368)
parser.add_argument("-r2", "--relation2_dim", type=int, default=64)
parser.add_argument("-r3", "--relation3_dim", type=int, default=8)
parser.add_argument("-w","--class_num",type = int, default=5)
parser.add_argument("-s","--sample_num_per_class",type = int, default=1)
parser.add_argument("-b","--batch_num_per_class",type = int, default=15)
parser.add_argument("-e","--episode",type = int, default= 500000)
parser.add_argument("-t","--test_episode", type = int, default=600)
parser.add_argument("-l","--learning_rate", type = float, default=0.001)
parser.add_argument("-u","--hidden_unit",type=int,default=10)
parser.add_argument("-tl", "--experiment_title", type=str, default="experiment-title")
args = parser.parse_args()


# Hyper Parameters
RELATION1_DIM = args.relation1_dim
RELATION2_DIM = args.relation2_dim
RELATION3_DIM = args.relation3_dim
CLASS_NUM = args.class_num
SAMPLE_NUM_PER_CLASS = 1
BATCH_NUM_PER_CLASS = args.batch_num_per_class
EPISODE = args.episode
TEST_EPISODE = args.test_episode
LEARNING_RATE = args.learning_rate
HIDDEN_UNIT = args.hidden_unit
EXPREIMENT_TITLE = args.experiment_title

x = torch.device("cuda")


def mean_confidence_interval(data, confidence=0.95):
    a = 1.0*np.array(data)
    n = len(a)
    m, se = np.mean(a), scipy.stats.sem(a)
    h = se * sp.stats.t._ppf((1+confidence)/2., n-1)
    return m, h


class ReshapeNetwork(nn.Module):
    """docstring for RelationNetwork"""
    def __init__(self):
        super(ReshapeNetwork, self).__init__()
        self.layer = nn.AvgPool2d(2)

    def forward(self, x):
        out = self.layer(x)
        return out


class CNNEncoder(nn.Module):
    """docstring for ClassName"""
    def __init__(self):
        super(CNNEncoder, self).__init__()
        self.layer1 = nn.Sequential(
                        nn.Conv2d(3, 64, kernel_size=3, padding=0),
                        nn.BatchNorm2d(64, momentum=1, affine=True),
                        nn.ReLU(),
                        nn.MaxPool2d(2))
        self.layer2 = nn.Sequential(
                        nn.Conv2d(64, 64, kernel_size=3, padding=0),
                        nn.BatchNorm2d(64, momentum=1, affine=True),
                        nn.ReLU(),
                        nn.MaxPool2d(2))
        self.layer3 = nn.Sequential(
                        nn.Conv2d(64, 64, kernel_size=3, padding=1),
                        nn.BatchNorm2d(64, momentum=1, affine=True),
                        nn.ReLU())
        self.layer4 = nn.Sequential(
                        nn.Conv2d(64, 64, kernel_size=3, padding=1),
                        nn.BatchNorm2d(64, momentum=1, affine=True),
                        nn.ReLU())

    def forward(self, x):
        out1 = self.layer1(x)
        # print('out1.size', out1.size())
        out2 = self.layer2(out1)
        # print('out2.size', out2.size())
        out3 = self.layer3(out2)
        # print('out3.size', out3.size())
        out4 = self.layer4(out3)
        # print('out4.size', out4.size())
        return out3, out4


class RelationNetwork(nn.Module):
    """docstring for RelationNetwork"""
    def __init__(self, input_size, hidden_size, output_size):
        super(RelationNetwork, self).__init__()
        self.fc1 = nn.Linear(input_size, hidden_size)
        self.fc2 = nn.Linear(hidden_size, output_size)
        self.fc3 = nn.Linear(output_size, 1)

    def forward(self, x):
        out = F.relu(self.fc1(x))
        out = F.relu(self.fc2(out))
        out = F.sigmoid(self.fc3(out))  # 注意，最后输出为（5*75，1）
        return out


def weights_init(m):
    classname = m.__class__.__name__
    if classname.find('Conv') != -1:
        n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
        m.weight.data.normal_(0, math.sqrt(2. / n))
        if m.bias is not None:
            m.bias.data.zero_()
    elif classname.find('BatchNorm') != -1:
        m.weight.data.fill_(1)
        m.bias.data.zero_()
    elif classname.find('Linear') != -1:
        n = m.weight.size(1)
        m.weight.data.normal_(0, 0.01)
        m.bias.data = torch.ones(m.bias.data.size())


def main():
    # Step 1: init data folders
    print("init data folders")
    metatrain_folders, metaval_folders = tg.mini_imagenet_folders()

    # Step 2: init neural networks
    print("init neural networks")
    reshapenetwork = ReshapeNetwork()
    feature_encoder = CNNEncoder()
    relation_network = RelationNetwork(RELATION1_DIM, RELATION2_DIM, RELATION3_DIM)

    feature_encoder.apply(weights_init)
    relation_network.apply(weights_init)

    reshapenetwork.to(x)
    feature_encoder.to(x)
    relation_network.to(x)

    feature_encoder_optim = torch.optim.Adam(feature_encoder.parameters(), lr=LEARNING_RATE)
    feature_encoder_scheduler = StepLR(feature_encoder_optim, step_size=100000, gamma=0.5)
    relation_network_optim = torch.optim.Adam(relation_network.parameters(), lr=LEARNING_RATE)
    relation_network_scheduler = StepLR(relation_network_optim, step_size=100000, gamma=0.5)

    saved_models_filepath, logs_filepath = build_experiment_folder(EXPREIMENT_TITLE)
    save_statistics(logs_filepath, ["episode", "val_accuracy", "loss"], create=True)

    if os.path.exists(str("{}/miniimagenet_feature_encoder.pkl".format(saved_models_filepath))):
        feature_encoder.load_state_dict(torch.load(
            str("{}/miniimagenet_feature_encoder.pkl".format(saved_models_filepath))))
        print("load feature encoder success")
    if os.path.exists(str("/{}/miniimagenet_relation_network.pkl".format(saved_models_filepath))):
        relation_network.load_state_dict(torch.load(
            str("{}/miniimagenet_relation_network.pkl".format(saved_models_filepath))))
        print("load relation network success")

    # Step 3: build graph
    print("Training...")
    last_accuracy = 0.0

    for episode in range(EPISODE):
        feature_encoder_scheduler.step(episode)
        relation_network_scheduler.step(episode)

        # init dataset
        task = tg.MiniImagenetTask(metatrain_folders, CLASS_NUM, SAMPLE_NUM_PER_CLASS, BATCH_NUM_PER_CLASS)
        sample_dataloader = tg.get_mini_imagenet_data_loader(task, num_per_class=SAMPLE_NUM_PER_CLASS,
                                                             split="train", shuffle=False)
        batch_dataloader = tg.get_mini_imagenet_data_loader(task, num_per_class=BATCH_NUM_PER_CLASS,
                                                            split="test", shuffle=True)

        # sample datas
        samples, sample_labels = sample_dataloader.__iter__().next()
        batches, batch_labels = batch_dataloader.__iter__().next()
        # samples.size():(5,3,84,84)
        # batches.size():(75,3,84,84)

        # calculate features
        sample_features3, sample_features4 = feature_encoder(Variable(samples).to(x))
        batch_features3, batch_features4 = feature_encoder(Variable(batches).to(x))

        sample_features3 = reshapenetwork(sample_features3.to(x))
        batch_features3 = reshapenetwork(batch_features3.to(x))

        sample_features4 = reshapenetwork(sample_features4.to(x))
        batch_features4 = reshapenetwork(batch_features4.to(x))

        sample_features = torch.cat((sample_features3, sample_features4), 1)
        # print("sample_features.shape", sample_features.shape)
        batch_features = torch.cat((batch_features3, batch_features4), 1)
        # print("batch_features.shape", batch_features.shape)

        # calculate relations
        sample_features_ext = sample_features.unsqueeze(0).repeat(BATCH_NUM_PER_CLASS*CLASS_NUM, 1, 1, 1, 1)
        # sample_features_ext.shape: (75,5,128,9,9)
        batch_features_ext = batch_features.unsqueeze(0).repeat(SAMPLE_NUM_PER_CLASS*CLASS_NUM, 1, 1, 1, 1)
        batch_features_ext = torch.transpose(batch_features_ext, 0, 1)
        # batch_features_ext.shape:(75,5,128,9,9)
        relation_pairs = torch.abs((sample_features_ext - batch_features_ext)).view(-1, 10368)
        # print(relation_pairs.shape)
        # relation_pairs.shape:(375,10368)
        relations = relation_network(relation_pairs).view(-1, CLASS_NUM*SAMPLE_NUM_PER_CLASS)

        mse = nn.MSELoss().to(x)
        change = batch_labels.view(-1, 1).long()
        one_hot_labels = Variable(torch.zeros(BATCH_NUM_PER_CLASS * CLASS_NUM, CLASS_NUM).scatter_(1, change, 1)).to(x)
        loss = mse(relations, one_hot_labels)

        # training
        feature_encoder.zero_grad()
        relation_network.zero_grad()

        loss.backward()

        torch.nn.utils.clip_grad_norm(feature_encoder.parameters(), 0.5)
        torch.nn.utils.clip_grad_norm(relation_network.parameters(), 0.5)

        feature_encoder_optim.step()
        relation_network_optim.step()

        if (episode+1) % 1000 == 0:
                print("episode:", episode+1, "loss", loss.item())

        if episode == 0 or (episode+1) % 5000 == 0:
            batch_size = batch_labels.shape[0]
            _, predict_labels = torch.max(relations.data, 1)
            rewards = [1 if predict_labels[j] == batch_labels[j] else 0 for j in range(batch_size)]
            total_rewards += np.sum(rewards)
            counter += batch_size
            accuracy = total_rewards / 1.0 / counter
            test_accuracy, h = mean_confidence_interval(accuracies)
            print("test accuracy:", test_accuracy, "h:", h)

            # test
            print("Testing...")
            accuracies = []
            for i in range(TEST_EPISODE):
                total_rewards = 0
                counter = 0
                num_per_class = 3
                task = tg.MiniImagenetTask(metaval_folders, CLASS_NUM, 1, num_per_class)
                sample_dataloader = tg.get_mini_imagenet_data_loader(task, num_per_class=1, split="train", shuffle=False)
                test_dataloader = tg.get_mini_imagenet_data_loader(task, num_per_class=num_per_class, split="test", shuffle=True)
                sample_images, sample_labels = sample_dataloader.__iter__().next()
                # print('sample_images',sample_images.size()) [5, 3, 84, 84])
                for test_images, test_labels in test_dataloader:
                    # print('test_images', test_images.size())([15, 3, 84, 84])
                    test_labels = test_labels.long().to(x)
                    batch_size = test_labels.shape[0]
                    # calculate features
                    sample_features3, sample_features4 = feature_encoder(Variable(sample_images).to(x))
                    test_features3, test_features4 = feature_encoder(Variable(test_images).to(x))

                    # calculate relations
                    sample_features3 = reshapenetwork(sample_features3.to(x))
                    test_features3 = reshapenetwork(test_features3.to(x))
                    sample_features4 = reshapenetwork(sample_features4.to(x))
                    test_features4 = reshapenetwork(test_features4.to(x))

                    sample_features = torch.cat((sample_features3, sample_features4), 1)
                    # print("sample_features ", sample_features.size()) ([5, 128, 9, 9])
                    test_features = torch.cat((test_features3, test_features4), 1)
                    # print("test_features", test_features.size()) ([15, 128, 9, 9])

                    # calculate relations
                    sample_features_ext = sample_features.unsqueeze(0).repeat(num_per_class * CLASS_NUM, 1, 1, 1, 1)
                    # print("sample_features_ext ", sample_features_ext.size()) ([5, 15, 128, 9, 9])
                    test_features_ext = test_features.unsqueeze(0).repeat(SAMPLE_NUM_PER_CLASS* CLASS_NUM, 1, 1, 1, 1)
                    # print("test_features_ext ", test_features_ext.size()) ([15, 5, 128, 9, 9])
                    test_features_ext = torch.transpose(test_features_ext, 0, 1)
                    relation_pairs = torch.abs((sample_features_ext - test_features_ext)).view(-1, 10368)
                    # print(relation_pairs.shape)  (75,10368)
                    relations = relation_network(relation_pairs).view(-1, CLASS_NUM)

                    _, predict_labels = torch.max(relations.data, 1)

                    rewards = [1 if predict_labels[j] == test_labels[j] else 0 for j in range(batch_size)]

                    total_rewards += np.sum(rewards)
                    counter += batch_size
                accuracy = total_rewards/1.0/counter
                accuracies.append(accuracy)

            test_accuracy, h = mean_confidence_interval(accuracies)

            print("test accuracy:", test_accuracy, "h:", h)
            save_statistics(logs_filepath, [episode, test_accuracy, loss.item()])

            # save networks
            if test_accuracy > last_accuracy:
                torch.save(feature_encoder.state_dict(),
                           str("{}/miniimagenet_feature_encoder.pkl".format(saved_models_filepath)))
                torch.save(relation_network.state_dict(),
                           str("{}/miniimagenet_relation_network.pkl".format(saved_models_filepath)))
                print("save networks for episode:", episode)
                last_accuracy = test_accuracy


if __name__ == '__main__':
    main()
