import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Variable
import numpy as np
import task_generator as tg
import math
import argparse
import random
import scipy as sp
import scipy.stats
from storage import save_statistics, build_experiment_folder

parser = argparse.ArgumentParser(description="One Shot Visual Recognition")
parser.add_argument("-r1", "--relation1_dim", type=int, default=512)
parser.add_argument("-r2", "--relation2_dim", type=int, default=128)
parser.add_argument("-r3", "--relation3_dim", type=int, default=32)
parser.add_argument("-r4", "--relation4_dim", type=int, default=8)
parser.add_argument("-c", "--class_num", type=int, default=5)
parser.add_argument("-s", "--sample_num_per_class", type=int, default=1)
parser.add_argument("-b", "--batch_num_per_class", type=int, default=19)
parser.add_argument("-e", "--episode", type=int, default=10)
parser.add_argument("-t", "--test_episode", type=int, default=1000)
parser.add_argument("-l", "--learning_rate", type=float, default=0.002)
parser.add_argument("-u", "--hidden_unit", type=int, default=10)
parser.add_argument("-tl", "--experiment_title", type=str, default="experiment-title")
args = parser.parse_args()


# Hyper Parameters
RELATION1_DIM = args.relation1_dim
RELATION2_DIM = args.relation2_dim
RELATION3_DIM = args.relation3_dim
RELATION4_DIM = args.relation4_dim
CLASS_NUM = args.class_num
SAMPLE_NUM_PER_CLASS = 1
BATCH_NUM_PER_CLASS = args.batch_num_per_class
EPISODE = args.episode
TEST_EPISODE = args.test_episode
LEARNING_RATE = args.learning_rate
HIDDEN_UNIT = args.hidden_unit
EXPREIMENT_TITLE = args.experiment_title

x = torch.device("cuda")


def mean_confidence_interval(data, confidence=0.95):
    a = 1.0*np.array(data)
    n = len(a)
    m, se = np.mean(a), scipy.stats.sem(a)
    h = se * sp.stats.t._ppf((1+confidence)/2., n-1)
    return m, h


class ReshapeNetwork(nn.Module):
    """docstring for RelationNetwork"""
    def __init__(self):
        super(ReshapeNetwork, self).__init__()
        self.layer = nn.AvgPool2d(2)

    def forward(self, x):
        out = self.layer(x)
        return out


class CNNEncoder(nn.Module):
    """docstring for ClassName"""
    def __init__(self):
        super(CNNEncoder, self).__init__()
        self.layer1 = nn.Sequential(
                        nn.Conv2d(1, 64, kernel_size=3, padding=0),
                        nn.BatchNorm2d(64, momentum=1, affine=True),
                        nn.ReLU(),
                        nn.MaxPool2d(2))
        self.layer2 = nn.Sequential(
                        nn.Conv2d(64, 64, kernel_size=3, padding=0),
                        nn.BatchNorm2d(64, momentum=1, affine=True),
                        nn.ReLU(),
                        nn.MaxPool2d(2))
        self.layer3 = nn.Sequential(
                        nn.Conv2d(64, 64, kernel_size=3, padding=1),
                        nn.BatchNorm2d(64, momentum=1, affine=True),
                        nn.ReLU())
        self.layer4 = nn.Sequential(
                        nn.Conv2d(64, 64, kernel_size=3, padding=1),
                        nn.BatchNorm2d(64, momentum=1, affine=True),
                        nn.ReLU())

    def forward(self,x):
        out1 = self.layer1(x)
        out2 = self.layer2(out1)
        out3 = self.layer3(out2)
        out4 = self.layer4(out3)
        return out3, out4


class RelationNetwork(nn.Module):
    """docstring for RelationNetwork"""
    def __init__(self, input_size, hidden1_size, hidden2_size, output_size):
        super(RelationNetwork, self).__init__()
        self.fc1 = nn.Linear(input_size, hidden1_size)  # (512,128)
        self.fc2 = nn.Linear(hidden1_size, hidden2_size)  # (128,32)
        self.fc3 = nn.Linear(hidden2_size, output_size)  # (32,8)
        self.fc4 = nn.Linear(output_size, 1)  # (8,1)

    def forward(self, x):
        out = F.relu(self.fc1(x))
        # print('out1.size', out.size())
        out = F.relu(self.fc2(out))
        # print('out2.size', out.size())
        out = F.relu(self.fc3(out))
        # print('out2.size', out.size())
        out = F.sigmoid(self.fc4(out))  # 注意，最后输出为（5*95，1）
        # print('out3.size', out.size())
        return out


def weights_init(m):
    classname = m.__class__.__name__
    if classname.find('Conv') != -1:
        n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
        m.weight.data.normal_(0, math.sqrt(2. / n))
        if m.bias is not None:
            m.bias.data.zero_()
    elif classname.find('BatchNorm') != -1:
        m.weight.data.fill_(1)
        m.bias.data.zero_()
    elif classname.find('Linear') != -1:
        n = m.weight.size(1)
        m.weight.data.normal_(0, 0.01)
        m.bias.data = torch.ones(m.bias.data.size())


def main():
    # Step 1: init data folders
    print("init data folders")
    # init character folders for dataset construction
    metatrain_character_folders, metaval_character_folders, metatest_character_folders = tg.omniglot_character_folders()

    # Step 2: init neural networks
    print("init neural networks")
    reshapenetwork = ReshapeNetwork()
    feature_encoder = CNNEncoder()
    relation_network = RelationNetwork(RELATION1_DIM, RELATION2_DIM, RELATION3_DIM, RELATION4_DIM)

    saved_models_filepath, logs_filepath = build_experiment_folder(EXPREIMENT_TITLE)
    save_statistics(logs_filepath, ["episode", "test accuracy"])

    feature_encoder.load_state_dict(torch.load(
            str("{}/omniglot_feature_encoder.pkl".format(saved_models_filepath))))
    print("load feature encoder success")

    relation_network.load_state_dict(torch.load(
        str("{}/omniglot_relation_network.pkl".format(saved_models_filepath))))
    print("load relation network success")

    total_accuracy = 0.0
    for episode in range(EPISODE):
            # test
            print("Testing...")
            total_rewards = 0
            accuracies = []
            for i in range(TEST_EPISODE):
                degrees = random.choice([0, 90, 180, 270])
                task = tg.OmniglotTask(metatest_character_folders, CLASS_NUM, SAMPLE_NUM_PER_CLASS, SAMPLE_NUM_PER_CLASS)
                # 验证集
                # 支持集和目标集，其实用test和train表示不太合理，应当该成target_set和support_set.
                sample_dataloader = tg.get_data_loader(task, num_per_class=SAMPLE_NUM_PER_CLASS,
                                                       split="support_set", shuffle=False, rotation=degrees)
                test_dataloader = tg.get_data_loader(task, num_per_class=SAMPLE_NUM_PER_CLASS,
                                                     split="target_set", shuffle=True, rotation=degrees)

                sample_images, sample_labels = sample_dataloader.__iter__().next()
                test_images, test_labels = test_dataloader.__iter__().next()
                # 注意在这里的test_images取了5张

                # calculate features
                sample_features3, sample_features4 = feature_encoder(Variable(sample_images))
                test_features3, test_features4 = feature_encoder(Variable(test_images))

                # calculate relations

                sample_features3 = reshapenetwork(sample_features3)
                test_features3 = reshapenetwork(test_features3)

                sample_features4 = reshapenetwork(sample_features4)
                test_features4 = reshapenetwork(test_features4)

                sample_features = torch.cat((sample_features3, sample_features4), 1)
                sample_features = sample_features.view(sample_features.size(0), -1)

                test_features = torch.cat((test_features3, test_features4), 1)
                test_features = test_features.view(test_features.size(0), -1)

                sample_features_ext = sample_features.unsqueeze(0).repeat(SAMPLE_NUM_PER_CLASS * CLASS_NUM, 1, 1)
                test_features_ext = test_features.unsqueeze(0).repeat(SAMPLE_NUM_PER_CLASS * CLASS_NUM, 1, 1)
                test_features_ext = torch.transpose(test_features_ext, 0, 1)

                relation_pairs = torch.abs((sample_features_ext - test_features_ext)).view(-1, 512)
                # print('relation_pairs :', relation_pairs.size())
                relations = relation_network(relation_pairs).view(-1, CLASS_NUM)
                # （25，1）到（5，5）

                _, predict_labels = torch.max(relations.data, 1)
                # print(predict_labels)
                test_labels = test_labels.long()
                rewards = [1 if predict_labels[j] == test_labels[j] else 0 for j in range(CLASS_NUM)]

                total_rewards += np.sum(rewards)
                accuracy = np.sum(rewards)/1.0/CLASS_NUM/SAMPLE_NUM_PER_CLASS
                accuracies.append(accuracy)

            test_accuracy, h = mean_confidence_interval(accuracies)
            print("test accuracy:", test_accuracy, "h:", h)
            save_statistics(logs_filepath, [episode, test_accuracy])
            total_accuracy += test_accuracy
    aver_accuracy = total_accuracy/EPISODE
    print("aver_accuracy:",aver_accuracy)
    save_statistics(logs_filepath, ["aver_accuracy"])
    save_statistics(logs_filepath, [aver_accuracy])


if __name__ == '__main__':
    main()
