'''34层特征+拉直后相减'''
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Variable
from torch.optim.lr_scheduler import StepLR
import numpy as np
import task_generator as tg
import os
import math
import argparse
import random
from storage import save_statistics, build_experiment_folder


parser = argparse.ArgumentParser(description="One Shot Visual Recognition")
parser.add_argument("-r1", "--relation1_dim", type=int, default=512)
parser.add_argument("-r2", "--relation2_dim", type=int, default=128)
parser.add_argument("-r3", "--relation3_dim", type=int, default=32)
parser.add_argument("-r4", "--relation4_dim", type=int, default=8)
parser.add_argument("-c", "--class_num", type=int, default=5)
parser.add_argument("-s", "--sample_num_per_class", type=int, default=1)
parser.add_argument("-b", "--batch_num_per_class", type=int, default=19)
parser.add_argument("-e", "--episode", type=int, default=1000000)
parser.add_argument("-t", "--test_episode", type=int, default=500)
parser.add_argument("-l", "--learning_rate", type=float, default=0.001)
parser.add_argument("-u", "--hidden_unit", type=int, default=10)
parser.add_argument("-tl", "--experiment_title", type=str, default="experiment-title")
args = parser.parse_args()

# Hyper Parameters
RELATION1_DIM = args.relation1_dim
RELATION2_DIM = args.relation2_dim
RELATION3_DIM = args.relation3_dim
RELATION4_DIM = args.relation4_dim
CLASS_NUM = args.class_num
SAMPLE_NUM_PER_CLASS = 1
BATCH_NUM_PER_CLASS = args.batch_num_per_class
EPISODE = args.episode
TEST_EPISODE = args.test_episode
LEARNING_RATE = args.learning_rate
HIDDEN_UNIT = args.hidden_unit
EXPREIMENT_TITLE = args.experiment_title

x = torch.device("cuda")


class ReshapeNetwork(nn.Module):
    """docstring for RelationNetwork"""
    def __init__(self):
        super(ReshapeNetwork, self).__init__()
        self.layer = nn.AvgPool2d(2)

    def forward(self, x):
        out = self.layer(x)
        return out


class CNNEncoder(nn.Module):
    """docstring for ClassName"""
    def __init__(self):
        super(CNNEncoder, self).__init__()
        self.layer1 = nn.Sequential(
                        nn.Conv2d(1, 64, kernel_size=3, padding=0),
                        nn.BatchNorm2d(64, momentum=1, affine=True),     # nn.BatchNorm2d表示批标准化
                        nn.ReLU(),                                       # nn.MaxPool2d表示最大池化，一般设置三个参数
                        nn.MaxPool2d(2))                                 # nn.Conv2d表示卷积模块，一般设置五个参数
        self.layer2 = nn.Sequential(
                        nn.Conv2d(64, 64, kernel_size=3, padding=0),
                        nn.BatchNorm2d(64, momentum=1, affine=True),
                        nn.ReLU(),
                        nn.MaxPool2d(2))
        self.layer3 = nn.Sequential(
                        nn.Conv2d(64, 64, kernel_size=3, padding=1),
                        nn.BatchNorm2d(64, momentum=1, affine=True),
                        nn.ReLU())
        self.layer4 = nn.Sequential(
                        nn.Conv2d(64, 64, kernel_size=3, padding=1),
                        nn.BatchNorm2d(64, momentum=1, affine=True),
                        nn.ReLU())

    def forward(self, x):
        # x.Size([b, 1, 28, 28])
        out1 = self.layer1(x)
        # print('out1.size', out1.size())
        out2 = self.layer2(out1)
        # print('out2.size', out2.size())
        out3 = self.layer3(out2)
        # print('out3.size', out3.size())
        out4 = self.layer4(out3)
        # print('out4.size', out.size())                         # 原文第四层输出维度为(5,5,64),注意改进网络时保证维度匹配
        return out3, out4


class RelationNetwork(nn.Module):
    """docstring for RelationNetwork"""
    def __init__(self, input_size, hidden1_size, hidden2_size, output_size):
        super(RelationNetwork, self).__init__()
        self.fc1 = nn.Linear(input_size, hidden1_size)  # (512,128)
        self.fc2 = nn.Linear(hidden1_size, hidden2_size)  # (128,32)
        self.fc3 = nn.Linear(hidden2_size, output_size)  # (32,8)
        self.fc4 = nn.Linear(output_size, 1)  # (8,1)

    def forward(self, x):
        out = F.relu(self.fc1(x))
        # print('out1.size', out.size())
        out = F.relu(self.fc2(out))
        # print('out2.size', out.size())
        out = F.relu(self.fc3(out))
        # print('out2.size', out.size())
        out = F.sigmoid(self.fc4(out))  # 注意，最后输出为（5*95，1）
        # print('out3.size', out.size())
        return out


def weights_init(m):
    """使用apply()时，需要先定义一个参数初始化的函数。之后，定义自己的网络，得到网络模型，
    使用apply()函数，就可以分别对conv层和bn层进行参数初始化。"""
    classname = m.__class__.__name__                     # 得到网络层的名字，如ConvTranspose2d
    if classname.find('Conv') != -1:                     # 使用了find函数，如果不存在返回值为-1，所以让其不等于-1
        n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
        m.weight.data.normal_(0, math.sqrt(2. / n))
        if m.bias is not None:
            m.bias.data.zero_()
    elif classname.find('BatchNorm') != -1:
        m.weight.data.fill_(1)
        m.bias.data.zero_()
    elif classname.find('Linear') != -1:
        m.weight.data.normal_(0, 0.01)
        m.bias.data = torch.ones(m.bias.data.size())


def main():
    # Step 1: init data folders
    print("init data folders")
    metatrain_character_folders, metaval_character_folders, metatest_character_folders = tg.omniglot_character_folders()

    # Step 2: init neural networks
    print("init neural networks")
    reshapenetwork = ReshapeNetwork()
    feature_encoder = CNNEncoder()
    relation_network = RelationNetwork(RELATION1_DIM, RELATION2_DIM, RELATION3_DIM, RELATION4_DIM)
    # 运用apply()函数进行权重初始化
    feature_encoder.apply(weights_init)
    relation_network.apply(weights_init)

    reshapenetwork.to(x)
    feature_encoder.to(x)
    relation_network.to(x)

    """要构建一个优化器optimizer，你必须给它一个可进行迭代优化的包含了所有参数（所有的参数必须是变量s）的列表。
     然后，您可以指定程序优化特定的选项，例如学习速率，权重衰减等。然后一般还会定义学习率的变化策略，
     这里采用的是torch.optim.lr_scheduler模块的StepLR类，表示每隔step_size个epoch就将学习率降为原来的gamma倍。"""
    feature_encoder_optim = torch.optim.Adam(feature_encoder.parameters(), lr=LEARNING_RATE)
    feature_encoder_scheduler = StepLR(feature_encoder_optim, step_size=1000000, gamma=0.5)

    relation_network_optim = torch.optim.Adam(relation_network.parameters(), lr=LEARNING_RATE)
    relation_network_scheduler = StepLR(relation_network_optim, step_size=100000, gamma=0.5)

    saved_models_filepath, logs_filepath = build_experiment_folder(EXPREIMENT_TITLE)
    save_statistics(logs_filepath, ["episode", "val_accuracy", "loss"], create=True)

    if os.path.exists(str("{}/omniglot_feature_encoder.pkl".format(saved_models_filepath))):
        feature_encoder.load_state_dict(torch.load(
            str("{}/omniglot_feature_encoder.pkl".format(saved_models_filepath))))
        print("load feature encoder success")
    if os.path.exists(str("/{}/omniglot_relation_network.pkl".format(saved_models_filepath))):
        relation_network.load_state_dict(torch.load(
            str("{}/omniglot_relation_network.pkl".format(saved_models_filepath))))
        print("load relation network success")

    # Step 3: build graph
    print("Training...")
    last_accuracy = 0.0

    for episode in range(EPISODE):
        # 训练开始的时候需要先更新下学习率，这是因为我们前面制定了学习率的变化策略，所以在每个epoch开始时都要更新下
        feature_encoder_scheduler.step(episode)
        relation_network_scheduler.step(episode)

        # init dataset
        # sample_dataloader is to obtain previous samples for compare
        # batch_dataloader is to batch samples for training
        degrees = random.choice([0, 90, 180, 270])
        # 制作支持集和目标集
        task = tg.OmniglotTask(metatrain_character_folders, CLASS_NUM, SAMPLE_NUM_PER_CLASS, BATCH_NUM_PER_CLASS)
        sample_dataloader = tg.get_data_loader(task, num_per_class=SAMPLE_NUM_PER_CLASS,
                                               split="support_set", shuffle=False, rotation=degrees)

        batch_dataloader = tg.get_data_loader(task, num_per_class=BATCH_NUM_PER_CLASS,
                                              split="target_set", shuffle=True, rotation=degrees)

        # sample datas
        samples, sample_labels = sample_dataloader.__iter__().next()
        batches, batch_labels = batch_dataloader.__iter__().next()
        # samples.size: torch.Size([5, 1, 28, 28]);sample_labels.size: torch.Size([5])
        # batches.size: torch.Size([95, 1, 28, 28]);batches_labels.size: torch.Size([95])
        # print(batch_labels.view(-1, 1))

        # 提取特征
        sample_features3, sample_features4 = feature_encoder(Variable(samples).to(x))
        # sample_features: torch.Size([5, 64, 5, 5])
        batch_features3, batch_features4 = feature_encoder(Variable(batches).to(x))
        # batch_features: torch.Size([95, 64, 5, 5])

        sample_features3 = reshapenetwork(sample_features3.to(x))
        batch_features3 = reshapenetwork(batch_features3.to(x))
        # sample_features3:torch.Size([5, 64, 2, 2])
        # print('batch_features3.size', batch_features3.size())

        sample_features4 = reshapenetwork(sample_features4.to(x))
        batch_features4 = reshapenetwork(batch_features4.to(x))
        # print('sample_features4.size', sample_features4.size())
        # print('batch_features4.size', batch_features4.size())

        sample_features = torch.cat((sample_features3, sample_features4), 1)
        sample_features = sample_features.view(sample_features.size(0), -1)
        # print('sample_features.size', sample_features.size())

        batch_features = torch.cat((batch_features3, batch_features4), 1)
        batch_features = batch_features.view(batch_features.size(0), -1)
        # print('batch_features.size', batch_features.size())

        # 拼接向量，其中torch.squeeze() 这个函数主要对数据的维度进行压缩，去掉维数为1的的维度
        sample_features_ext = sample_features.unsqueeze(0).repeat(BATCH_NUM_PER_CLASS*CLASS_NUM, 1, 1)
        # sample_features_ext : torch.Size([95, 5, 64, 5, 5])
        batch_features_ext = batch_features.unsqueeze(0).repeat(SAMPLE_NUM_PER_CLASS*CLASS_NUM, 1, 1)
        # batch_features_ext: torch.Size([5, 95, 64, 5, 5])
        batch_features_ext = torch.transpose(batch_features_ext, 0, 1)
        # batch_features_ext after: torch.Size([95, 5, 64, 5, 5])
        # 在深度学习处理图像的时候，经常要考虑将多张不同图片输入到网络，这时需要用torch.cat([image1,image2],1),
        relation_pairs = torch.abs((sample_features_ext - batch_features_ext)).view(-1, 512)
        # 度量学习
        relations = relation_network(relation_pairs).view(-1, CLASS_NUM)
        # relations torch.Size([95, 5])

        # 优化目标
        mse = nn.MSELoss().to(x)

        change = batch_labels.view(-1, 1).long()
        one_hot_labels = Variable(torch.zeros(BATCH_NUM_PER_CLASS * CLASS_NUM, CLASS_NUM).scatter_(1, change, 1)).to(x)
        loss = mse(relations, one_hot_labels)
        # batch_labels = batch_labels.long()
        # loss = loss(relations, batch_labels)

        # training 然后先将网络中的所有梯度置0
        feature_encoder.zero_grad()
        relation_network.zero_grad()

        loss.backward()  # 计算得到loss后就要回传损失
        # 梯度剪裁
        torch.nn.utils.clip_grad_norm(feature_encoder.parameters(), 0.5)
        torch.nn.utils.clip_grad_norm(relation_network.parameters(), 0.5)
        # 回传损失过程中会计算梯度，然后需要根据这些梯度更新参数，XX.step()就是用来更新参数的。之后，
        # 你就可以从xx.param_groups[0][‘params’]里面看到各个层的梯度和权值信息。
        feature_encoder_optim.step()
        relation_network_optim.step()

        if (episode+1) % 1000 == 0:
                print("episode:", episode+1, "loss", loss.item())

        if episode == 0 or (episode+1) % 5000 == 0:
            # test
            print("Testing...")

            total_rewards = 0
            for i in range(TEST_EPISODE):
                degrees = random.choice([0, 90, 180, 270])
                task = tg.OmniglotTask(metaval_character_folders, CLASS_NUM,
                                       SAMPLE_NUM_PER_CLASS, SAMPLE_NUM_PER_CLASS,)
                sample_dataloader = tg.get_data_loader(task, num_per_class=SAMPLE_NUM_PER_CLASS,
                                                       split="support_set", shuffle=False, rotation=degrees)
                test_dataloader = tg.get_data_loader(task, num_per_class=SAMPLE_NUM_PER_CLASS,
                                                     split="target_set", shuffle=True, rotation=degrees)

                sample_images, sample_labels = sample_dataloader.__iter__().next()
                val_images, val_labels = test_dataloader.__iter__().next()
                val_labels = val_labels.long()
                # print('test_labels', test_labels)

                sample_features3, sample_features4 = feature_encoder(Variable(sample_images).to(x))
                batch_val_features3, batch_val_features4 = feature_encoder(Variable(val_images).to(x))

                sample_features3 = reshapenetwork(sample_features3.to(x))
                batch_val_features3 = reshapenetwork(batch_val_features3.to(x))

                sample_features4 = reshapenetwork(sample_features4.to(x))
                batch_val_features4 = reshapenetwork(batch_val_features4.to(x))

                sample_features = torch.cat((sample_features3, sample_features4), 1)
                sample_features = sample_features.view(sample_features.size(0), -1)
                # print('sample_features.size', sample_features.size())

                batch_features = torch.cat((batch_val_features3, batch_val_features4), 1)
                batch_features = batch_features.view(batch_features.size(0), -1)

                # calculate relations
                # each batch sample link to every samples to calculate relations
                # to form a 100x128 matrix for relation network
                sample_features_ext = sample_features.unsqueeze(0).repeat(SAMPLE_NUM_PER_CLASS * CLASS_NUM, 1, 1)
                val_features_ext = batch_features.unsqueeze(0).repeat(SAMPLE_NUM_PER_CLASS * CLASS_NUM, 1, 1)
                val_features_ext = torch.transpose(val_features_ext, 0, 1)

                relation_pairs = torch.abs((sample_features_ext - val_features_ext)).view(-1, 512)
                relations = relation_network(relation_pairs).view(-1, CLASS_NUM)

                _, predict_labels = torch.max(relations.data, 1)
                val_labels = val_labels.long().to(x)
                rewards = [1 if predict_labels[j] == val_labels[j] else 0 for j in range(CLASS_NUM)]
                # print('rewards', rewards)
                total_rewards += np.sum(rewards)

            val_accuracy = total_rewards/1.0/CLASS_NUM/TEST_EPISODE
            print("val accuracy:", val_accuracy)
            save_statistics(logs_filepath, [episode+1, val_accuracy, loss.item()])

            if val_accuracy > last_accuracy:
                # save networks
                torch.save(feature_encoder.state_dict(),
                           str("{}/omniglot_feature_encoder.pkl".format(saved_models_filepath)))
                torch.save(relation_network.state_dict(),
                           str("{}/omniglot_relation_network.pkl".format(saved_models_filepath)))
                print("save networks for episode:", episode+1)
                last_accuracy = val_accuracy


if __name__ == '__main__':
    main()
