import torchvision.transforms as transforms
import torch
from torch.utils.data import DataLoader, Dataset
import random
import os
from PIL import Image
import matplotlib.pyplot as plt
import numpy as np
from torch.utils.data.sampler import Sampler


def imshow(img):
    npimg = img.numpy()
    plt.axis("off")
    plt.imshow(np.transpose(npimg, (1, 2, 0)))
    plt.show()


class Rotate(object):
    def __init__(self, angle):
        self.angle = angle

    def __call__(self, x, mode="reflect"):           # 一个类实例也可以变成一个可调用对象，只需要实现一个特殊方法__call__()
        x = x.rotate(self.angle)
        return x


# 获取1623类图片集路径
def omniglot_character_folders():
    data_folder = '../datas/omniglot_resized/'
    character_folders = [os.path.join(data_folder, family, character) for family in os.listdir(data_folder)
                         if os.path.isdir(os.path.join(data_folder, family))
                         for character in os.listdir(os.path.join(data_folder, family))]
    random.seed(1)
    random.shuffle(character_folders)

    num_train = 1200
    num_val = 1411
    metatrain_character_folders = character_folders[:num_train]
    metaval_character_folders = character_folders[num_train:num_val]
    metatest_character_folders = character_folders[num_val:]

    return metatrain_character_folders, metaval_character_folders, metatest_character_folders


# 制作每个迭代任务目标集和支持集
class OmniglotTask(object):
    def __init__(self, character_folders, num_classes, support_set_num, target_set_num):
        self.character_folders = character_folders
        self.num_classes = num_classes
        self.support_set_num = support_set_num
        self.target_set_num = target_set_num
        self.support_set = []
        self.target_set = []

        class_folders = random.sample(self.character_folders, self.num_classes)  # 选取N类，eg:20
        labels = np.array(range(len(class_folders)))
        labels = dict(zip(class_folders, labels))
        samples = {}

        for c in class_folders:
            temp = [os.path.join(c, x) for x in os.listdir(c)]  # 每类20张图片的路径
            samples[c] = random.sample(temp, len(temp))   # {0：[20张图片的路径],1:[20张图片的路径],...4:[20张图片的路径]}

            self.support_set += samples[c][:support_set_num]           # 支持集每类选取k张，eg:5张,支持集含有25张
            self.target_set += samples[c][support_set_num:support_set_num + target_set_num]
            # 目标集每类选取n*batch_size张，eg:5*5=25

        self.support_set_labels = [labels[self.get_class(x)] for x in self.support_set]
        self.target_set_labels = [labels[self.get_class(x)] for x in self.target_set]

    def get_class(self, sample):
        return os.path.join(*sample.split('\\')[:-1])


# 是选取支持集还是目标集
class FewShotDataset(Dataset):
    def __init__(self, task, split='support_set', transform=None, target_transform=None):
        self.transform = transform  # Torch operations on the input image
        self.target_transform = target_transform
        self.task = task
        self.split = split
        self.image_roots = self.task.support_set if self.split == 'support_set' else self.task.target_set
        self.labels = self.task.support_set_labels if self.split == 'support_set' else self.task.target_set_labels

    def __len__(self):
        return len(self.image_roots)

    def __getitem__(self, idx):
        raise NotImplementedError("This is an abstract class. Subclass this class for your particular dataset.")


# 图片处理，旋转与读取
class Omniglot(FewShotDataset):

    def __init__(self, *args, **kwargs):
        super(Omniglot, self).__init__(*args, **kwargs)

    def __getitem__(self, idx):
        image_root = self.image_roots[idx]
        image = Image.open(image_root)
        image = image.convert('L')      # 模式L为灰色图像
        image = image.resize((28, 28), resample=Image.LANCZOS)  # per Chelsea's implementation
        # image = np.array(image, dtype=np.float32)
        if self.transform is not None:
            image = self.transform(image)

        label = self.labels[idx]

        if self.target_transform is not None:
            label = self.target_transform(label)
        return image, label


# 生成迭代器
class ClassBalancedSampler(Sampler):
    def __init__(self, num_per_class, num_cl, num_inst, shuffle=True):
        self.num_per_class = num_per_class
        self.num_cl = num_cl
        self.num_inst = num_inst
        self.shuffle = shuffle

    def __iter__(self):
        # return a single list of indices, assuming that items will be grouped by class
        if self.shuffle:      # torch.randperm返回一个0到n-1的数组
            batch = [[i+j*self.num_inst for i in torch.randperm(self.num_inst)[:self.num_per_class]]
                     for j in range(self.num_cl)]
        else:
            batch = [[i+j*self.num_inst for i in range(self.num_inst)[:self.num_per_class]] for j in range(self.num_cl)]
        batch = [item for sublist in batch for item in sublist]

        if self.shuffle:
            random.shuffle(batch)
        return iter(batch)

    def __len__(self):
        return 1


# 加载数据
def get_data_loader(task, num_per_class=1, split='support_set', shuffle=True, rotation=0):
    # NOTE: batch size here is # instances PER CLASS
    normalize = transforms.Normalize(mean=[0.92206, 0.92206, 0.92206], std=[0.08426, 0.08426, 0.08426])

    dataset = Omniglot(task, split=split,
                       transform=transforms.Compose([Rotate(rotation), transforms.ToTensor(), normalize]))

    if split == 'support_set':
        sampler = ClassBalancedSampler(num_per_class, task.num_classes, task.support_set_num, shuffle=shuffle)
    else:
        sampler = ClassBalancedSampler(num_per_class, task.num_classes, task.target_set_num, shuffle=shuffle)

    loader = DataLoader(dataset, batch_size=num_per_class*task.num_classes, sampler=sampler)

    return loader

